#
# Helper functions for boxes.
#

import numpy as np
from scipy.stats import entropy
from scipy.spatial.distance import cosine

from sklearn.metrics import average_precision_score

#
# Simple NMS implementation.
#
def nms(dets, thresh):
    x1 = dets[:, 0]
    y1 = dets[:, 1]
    x2 = dets[:, 2]
    y2 = dets[:, 3]
    scores = dets[:, 4]

    areas = (x2 - x1 + 1) * (y2 - y1 + 1)
    order = scores.argsort()[::-1]

    keep = []
    while order.size > 0:
        i = order[0]
        keep.append(i)
        xx1 = np.maximum(x1[i], x1[order[1:]])
        yy1 = np.maximum(y1[i], y1[order[1:]])
        xx2 = np.minimum(x2[i], x2[order[1:]])
        yy2 = np.minimum(y2[i], y2[order[1:]])

        w = np.maximum(0.0, xx2 - xx1 + 1)
        h = np.maximum(0.0, yy2 - yy1 + 1)
        inter = w * h
        ovr = inter / (areas[i] + areas[order[1:]] - inter)

        inds = np.where(ovr <= thresh)[0]
        order = order[inds + 1]

    return keep

#
#
#
def object_object_match(owm):
    cs = np.zeros((owm.shape[0], owm.shape[0]))
    for i in range(owm.shape[0]):
        for j in range(i, owm.shape[0]):
            cs[i,j] = cs[j,i] = 1 - cosine(owm[i], owm[j])
    return cs

#
#
#
def action_object_scores(awm, owm, action):
    # Compute similarities.
    cs = np.zeros(owm[0].shape[0])
    for i in range(len(awm)):
        a, b = awm[i], owm[i]
        for k in range(b.shape[0]):
            cs[k] += (1 - cosine(a[action], b[k]))
        cs[np.isnan(cs)] = -10
    return cs / len(awm)

#
#
#
def action_object_match(awm, owm, action, topn, use_od, use_ad, weights, oom):
    # Compute similarities.
    acs = np.zeros((awm[0].shape[0], owm[0].shape[0]))
    for i in range(len(awm)):
        a, b = awm[i], owm[i]
        for j in range(a.shape[0]):
            for k in range(b.shape[0]):
                acs[j,k] += (1 - cosine(a[j], b[k]))
        acs[np.isnan(acs)] = -10
    #cs /= len(awm)
    #acs[np.isnan(acs)] = -10
    
    # Incorporate object weights.
    cs  = acs[action]
    cs /= len(awm)
    wcs = acs[action] * weights
    
    # Object-based discrimination.
    if use_od:
        entropies = cs + (np.mean(oom, axis=0) ** 0.5)
        wcs += entropies
    
    # Action-based discrimination.
    if use_ad:
        actions = np.arange(awm[0].shape[0])
        other_actions = np.setdiff1d(actions, action)
        entropies = cs - np.max(acs[other_actions,:], axis=0)
        wcs += entropies
    
    # Find top objects.
    co = np.argsort(wcs)[::-1]
    ci = co[:topn]
    cw = wcs[co[:topn]]
    return ci, cw

#
# Box intersection.
#
def boxintersect(a,b):
    if a[0] > b[2] or b[0] > a[2] or a[1] > b[3] or b[1] > a[3]:
        return 0
    w = min(a[2], b[2]) - max(a[0], b[0])
    h = min(a[3], b[3]) - max(a[1], b[1])
    return w * h

#
# Minimal edge distance between two boxes.
#
def box_dist(a, b):
    if boxintersect(a,b) > 0:
        return 0
    ae = np.array([[a[0],a[1]], [a[2],a[1]], [a[0],a[3]], [a[2],a[3]]])
    be = np.array([[b[0],b[1]], [b[2],b[1]], [b[0],b[3]], [b[2],b[3]]])
    mind = np.min(np.linalg.norm(ae-be[0], axis=1))
    for i in range(1, be.shape[0]):
        nd   = np.min(np.linalg.norm(ae-be[i], axis=1))
        mind = min(mind, nd)
    return mind

#
# Tile distributions with 9 tiles. a=person, b=object.
#
def tiledist(a, b):
    d = np.zeros(9)
    e = 1e6
    # Above left.
    d[0] = boxintersect([0, 0, a[0], a[1]], b)
    # Above center.
    d[1] = boxintersect([a[0], 0, a[2], a[1]], b)
    # Above right.
    d[2] = boxintersect([a[2], 0, e, a[1]], b)
    # Left.
    d[3] = boxintersect([0, a[1], a[0], a[3]], b)
    # On.
    d[4] = boxintersect(a, b)
    # Right.
    d[5] = boxintersect([a[2], a[1], e, a[3]], b)
    # Below left.
    d[6] = boxintersect([0, a[3], a[0], e], b)
    # Below center.
    d[7] = boxintersect([a[0], a[3], a[2], e], b)
    # Below right.
    d[8] = boxintersect([a[2], a[3], e, e], b)
    return d / float((b[2] - b[0]) * (b[3] - b[1]))

#
# Tile distributions with 16 tiles. a=person, b=object.
#
def tiledist_four(a, b):
    d = np.zeros(16)
    e = 1e6

    d[0] = boxintersect([0, 0, a[0], a[1]], b)
    d[1] = boxintersect([a[0], 0, (a[0]+a[2])/2., a[1]], b)
    d[2] = boxintersect([(a[0]+a[2])/2., 0, a[2], a[1]], b)
    d[3] = boxintersect([a[2], 0, e, a[1]], b)
    
    d[4] = boxintersect([0, a[1], a[0], (a[1]+a[3])/2.], b)
    d[5] = boxintersect([a[0], a[1], (a[0]+a[2])/2., (a[1]+a[3])/2.], b)
    d[6] = boxintersect([(a[0]+a[2])/2., a[1], a[2], (a[1]+a[3])/2.], b)
    d[7] = boxintersect([a[2], a[1], e, (a[1]+a[3])/2.], b)
    
    d[8] = boxintersect([0, (a[1]+a[3])/2., a[0], a[3]], b)
    d[9] = boxintersect([a[0], (a[1]+a[3])/2., (a[0]+a[2])/2., a[3]], b)
    d[10] = boxintersect([(a[0]+a[2])/2, (a[1]+a[3])/2., a[2], a[3]], b)
    d[11] = boxintersect([a[2], (a[1]+a[3])/2., e, a[3]], b)
    
    d[12] = boxintersect([0, a[3], a[0], e], b)
    d[13] = boxintersect([a[0], a[3], (a[0]+a[2])/2., e], b)
    d[14] = boxintersect([(a[0]+a[2])/2, a[3], a[2], e], b)
    d[15] = boxintersect([a[2], a[3], e, e], b)

    return d / float((b[2] - b[0]) * (b[3] - b[1]))

#
# Tile distributions with 16 tiles. a=person, b=object.
#
def tiledist_five(a, b):
    d = np.zeros(25)
    e = 1e6

    ax1, ax2 = (2*a[0]+a[2]) / 3., (a[0] + 2*a[2])/3.
    ay1, ay2 = (2*a[1]+a[3]) / 3., (a[1] + 2*a[3])/3.

    d[0] = boxintersect([0, 0, a[0], a[1]], b)
    d[1] = boxintersect([a[0], 0, ax1, a[1]], b)
    d[2] = boxintersect([ax1, 0, ax2, a[1]], b)
    d[3] = boxintersect([ax2, 0, a[2], a[1]], b)
    d[4] = boxintersect([a[2], 0, e, a[1]], b)

    d[5] = boxintersect([0, a[1], a[0], ay1], b)
    d[6] = boxintersect([a[0], a[1], ax1, ay1], b)
    d[7] = boxintersect([ax1, a[1], ax2, ay1], b)
    d[8] = boxintersect([ax2, a[1], a[2], ay1], b)
    d[9] = boxintersect([a[2], a[1], e, ay1], b)

    d[10] = boxintersect([0, ay1, a[0], ay2], b)
    d[11] = boxintersect([a[0], ay1, ax1, ay2], b)
    d[12] = boxintersect([ax1, ay1, ax2, ay2], b)
    d[13] = boxintersect([ax2, ay1, a[2], ay2], b)
    d[14] = boxintersect([a[2], ay1, e, ay2], b)

    d[15] = boxintersect([0, ay2, a[0], a[3]], b)
    d[16] = boxintersect([a[0], ay2, ax1, a[3]], b)
    d[17] = boxintersect([ax1, ay2, ax2, a[3]], b)
    d[18] = boxintersect([ax2, ay2, a[2], a[3]], b)
    d[19] = boxintersect([a[2], ay2, e, a[3]], b)

    d[20] = boxintersect([0, a[3], a[0], e], b)
    d[21] = boxintersect([a[0], a[3], ax1, e], b)
    d[22] = boxintersect([ax1, a[3], ax2, e], b)
    d[23] = boxintersect([ax2, a[3], a[2], e], b)
    d[24] = boxintersect([a[2], a[3], e, e], b)

    return d / float((b[2] - b[0]) * (b[3] - b[1]))

#
# Overlap between two boxes.
#
def box_overlap(a, b):
    if a[2] > b[0] and b[2] > a[0] and a[3] > b[1] and b[3] > a[1]:
        i  = min(a[2],b[2]) - max(a[0],b[0])
        i *= min(a[3],b[3]) - max(a[1],b[1])
        i  = float(i)
        a1 = ((a[2]-a[0]) * (a[3]-a[1]))
        a2 = ((b[2]-b[0]) * (b[3]-b[1]))
        return i / (a1 + a2 - i)
    return 0.

#
# Overlap between box and list of other boxes.
#
def liou(a, b):
    iou = np.zeros(b.shape[0])
    for i in range(b.shape[0]):
        iou[i] = box_overlap(a, b[i])
    return iou
    
#
# Jensen-Shannon Divergence.
#
def jsd(p, q):
    apq = 0.5 * (p + q)
    return 0.5 * entropy(p, apq, 2) + 0.5 * entropy(q, apq, 2)

#
#
#
def tube_trim(tube):
    keep = np.where(tube[:,1] >= 0)[0]
    return tube[keep,:]

#
# Perform interpolation (for UCF-101 and H2T) of generated tubes.
#
def tube_interpolate(tube, scores, stride, nr_frames):
    if tube.shape[0] == nr_frames:
        return tube
    
    ntube   = np.zeros((nr_frames, tube.shape[1]), dtype=tube.dtype)
    nscores = np.zeros(nr_frames)

    for i in range(nr_frames):
        i1, i2 = i / stride, i / stride + 1
        w      = (i % stride) / float(stride)

        ntube[i,0] = i
        if i2 < tube.shape[0]:
            ntube[i,1] = (1-w) * tube[i1,1] + w * tube[i2,1]
            ntube[i,2] = (1-w) * tube[i1,2] + w * tube[i2,2]
            ntube[i,3] = (1-w) * tube[i1,3] + w * tube[i2,3]
            ntube[i,4] = (1-w) * tube[i1,4] + w * tube[i2,4]
            nscores[i] = (1-w) * scores[i1] + w * nscores[i2]
        else:
            ntube[i,1] = tube[i1,1]
            ntube[i,2] = tube[i1,2]
            ntube[i,3] = tube[i1,3]
            ntube[i,4] = tube[i1,4]
            nscores[i] = scores[i1]

    return ntube, nscores
    
#
# Compute the intersection-over-union score for two proposals from the
# same video.
#
def tube_iou(p1, p2, ss=1, si=-1):
    # Frame indices.
    p2idxs   = np.where(p2[:,2] >= 0)[0]
    p1       = p1[::ss,:]
    if si > 0:
        p1 = p1[si:-si,:]
    p1f, p2f = p1[:,0].astype(int), p2[:,0].astype(int)
    p2f      = p2f[p2idxs]
    p2       = p2[p2idxs,:]

    # Determine union of frame span.
    tmin = min(np.min(p1f), np.min(p2f))
    tmax = max(np.max(p1f), np.max(p2f))

    # Initialize the overlap scores across frame span.
    span = np.arange(tmin, tmax+1, ss)
    overlaps  = np.zeros(len(span), dtype=np.float)

    # Go through the frame span.
    for d in range(len(span)):
        i = span[d]
        p1i, p2i = np.where(p1f == i)[0], np.where(p2f == i)[0]
        # Compute the overlap if frame in both proposals.
        if len(p1i) == 1 and len(p2i) == 1:
            a,b = p1[p1i[0],1:], p2[p2i[0],1:]
            a   = [min(a[0],a[2]), min(a[1],a[3]), max(a[0],a[2]), max(a[1],a[3])]
            b   = [min(b[0],b[2]), min(b[1],b[3]), max(b[0],b[2]), max(b[1],b[3])]
            # Only compute overlap if there is any
            if a[2] > b[0] and b[2] > a[0] and a[3] > b[1] and b[3] > a[1]:
                intersection  = (min(a[2],b[2]) - max(a[0],b[0]))
                intersection *= (min(a[3],b[3]) - max(a[1],b[1]))
                intersection  = float(intersection)
                area1         = ((a[2]-a[0]) * (a[3]-a[1]))
                area2         = ((b[2]-b[0]) * (b[3]-b[1]))
                overlaps[d] = intersection / (area1 + area2 - intersection)

    # Return the mean overlap over the frame span
    return np.mean(overlaps)

#
# Compute localization APs.
#
def compute_ap(overlaps, labels, predictions, actioncounter, iout, ignore_dup=True):
    # Initialize container.
    aps = np.zeros((overlaps.shape[0], len(iout)))
    
    # Go over all action classifiers and thresholds.
    for i in range(overlaps.shape[0]):
        for j in range(len(iout)):
            # Binarize the labels.
            testy  = overlaps[i,:,:] >= iout[j]
            testy *= labels[i,:,:]
            testy  = testy.astype(int)
            
            # Find duplicates, and ignore or penalize them..
            for k in range(testy.shape[0]):
                fidx = -1
                for l in range(testy.shape[1]):
                    if testy[k,l] == 1:
                        if fidx == -1:
                            fidx = l
                        else:
                            if ignore_dup:
                                testy[k,l] = 2
                            else:
                                testy[k,l] = 2

            # Ignore or penalize duplicates.
            #if ignore_dup:
            oidxs = np.where(testy != 2)
            testy = testy[oidxs[0], oidxs[1]].ravel()
            predy = predictions[i, oidxs[0], oidxs[1]].ravel()
            #else:
            #    testy[testy == 2] = 0
            #    testy = testy.ravel()
            #    predy = self.predictions[i,:,:].ravel()

            # Compute the Average Precision.
            if np.sum(testy == 1) > 0:
                ap = average_precision_score(testy, predy)
                aps[i,j] = ap * np.sum(testy == 1) / actioncounter[i]
    
    return aps

#
# Compute localization AUCs.
#
def compute_auc(overlaps, labels, predictions, actioncounter, iout, ignore_dup=True):
    aucs = np.zeros(len(iout))

    for i in range(len(iout)):
        testy  = overlaps[:,:,:] >= iout[i]
        testy *= labels[:,:,:]
        testy  = testy.astype(int)

        for j in range(testy.shape[0]):
            for k in range(testy.shape[1]):
                fidx = -1
                for l in range(testy.shape[2]):
                    if testy[j,k,l] == 1:
                        if fidx == -1:
                            fidx = l
                        else:
                            if ignore_dup:
                                testy[j,k,l] = 2
                            else:
                                testy[j,k,l] = 0

        ns = (testy.shape[0], testy.shape[1] * testy.shape[2])
        testy = np.reshape(testy, ns)
        predy = np.reshape(predictions.copy(), ns)

        mins = np.min(predy, axis=1)
        maxs = np.max(predy, axis=1)
        mins = np.tile(mins, (predy.shape[1],1)).T
        maxs = np.tile(maxs, (predy.shape[1],1)).T
        predy -= mins
        predy /= (maxs - mins)

        testy = testy.ravel()
        predy = predy.ravel()
        nidxs = np.where(testy != 2)[0]
        testy = testy[nidxs]
        predy = predy[nidxs]

        order = np.argsort(predy)[::-1]
        tp = np.cumsum(testy[order] == 1)
        fp = np.cumsum(testy[order] == 0)

        rec = tp / float(np.sum(actioncounter))
        fpr = fp / float(np.sum(testy == 0))

        uidxs = np.where(fpr <= 0.6)[0]
        aucs[i] = np.abs(np.trapz(rec[uidxs], fpr[uidxs]))
    
    return aucs
